@@ -1,8 +1,8 @@
 // Scintilla source code edit control
 /** @file LexHTML.cxx
  ** Lexer for HTML.
- **/ 
-// Copyright 1998-2002 by Neil Hodgson <neilh@scintilla.org>
+ **/
+// Copyright 1998-2003 by Neil Hodgson <neilh@scintilla.org>
 // The License.txt file describes the conditions under which this software may be distributed.
 
 #include <stdlib.h>
@@ -24,24 +24,35 @@
 #define SCE_HA_VBS (SCE_HBA_START - SCE_HB_START)
 #define SCE_HA_PYTHON (SCE_HPA_START - SCE_HP_START)
 
-enum { eScriptNone = 0, eScriptJS, eScriptVBS, eScriptPython, eScriptPHP, eScriptXML, eScriptSGML, eScriptSGMLblock };
-enum { eHtml = 0, eNonHtmlScript, eNonHtmlPreProc, eNonHtmlScriptPreProc };
+enum script_type { eScriptNone = 0, eScriptJS, eScriptVBS, eScriptPython, eScriptPHP, eScriptXML, eScriptSGML, eScriptSGMLblock };
+enum script_mode { eHtml = 0, eNonHtmlScript, eNonHtmlPreProc, eNonHtmlScriptPreProc };
 
-inline bool IsAWordChar(const int ch) {
+static inline bool IsAWordChar(const int ch) {
 	return (ch < 0x80) && (isalnum(ch) || ch == '.' || ch == '_');
 }
 
-inline bool IsAWordStart(const int ch) {
+static inline bool IsAWordStart(const int ch) {
 	return (ch < 0x80) && (isalnum(ch) || ch == '_');
 }
 
-static int segIsScriptingIndicator(Accessor &styler, unsigned int start, unsigned int end, int prevValue) {
-	char s[30 + 1];
-	unsigned int i = 0;
-	for (; i < end - start + 1 && i < 30; i++) {
-		s[i] = static_cast<char>(tolower(styler[start + i]));
+static inline int MakeLowerCase(int ch) {
+	if (ch < 'A' || ch > 'Z')
+		return ch;
+	else
+		return ch - 'A' + 'a';
+}
+
+static void GetTextSegment(Accessor &styler, unsigned int start, unsigned int end, char *s, size_t len) {
+	size_t i = 0;
+	for (; (i < end - start + 1) && (i < len-1); i++) {
+		s[i] = static_cast<char>(MakeLowerCase(styler[start + i]));
 	}
 	s[i] = '\0';
+}
+
+static script_type segIsScriptingIndicator(Accessor &styler, unsigned int start, unsigned int end, script_type prevValue) {
+	char s[100];
+	GetTextSegment(styler, start, end, s, sizeof(s));
 	//Platform::DebugPrintf("Scripting indicator [%s]\n", s);
 	if (strstr(s, "src"))	// External script
 		return eScriptNone;
@@ -63,12 +74,8 @@ static int segIsScriptingIndicator(Accessor &styler, unsigned int start, unsigne
 
 static int PrintScriptingIndicatorOffset(Accessor &styler, unsigned int start, unsigned int end) {
 	int iResult = 0;
-	char s[30 + 1];
-	unsigned int i = 0;
-	for (; i < end - start + 1 && i < 30; i++) {
-		s[i] = static_cast<char>(tolower(styler[start + i]));
-	}
-	s[i] = '\0';
+	char s[100];
+	GetTextSegment(styler, start, end, s, sizeof(s));
 	if (0 == strncmp(s, "php", 3)) {
 		iResult = 3;
 	}
@@ -76,7 +83,7 @@ static int PrintScriptingIndicatorOffset(Accessor &styler, unsigned int start, u
 	return iResult;
 }
 
-static int ScriptOfState(int state) {
+static script_type ScriptOfState(int state) {
 	if ((state >= SCE_HP_START) && (state <= SCE_HP_IDENTIFIER)) {
 		return eScriptPython;
 	} else if ((state >= SCE_HB_START) && (state <= SCE_HB_STRINGEOL)) {
@@ -85,7 +92,7 @@ static int ScriptOfState(int state) {
 		return eScriptJS;
 	} else if ((state >= SCE_HPHP_DEFAULT) && (state <= SCE_HPHP_COMMENTLINE)) {
 		return eScriptPHP;
-	} else if ((state >= SCE_H_SGML_DEFAULT) && (state <= SCE_H_SGML_BLOCK_DEFAULT)) {
+	} else if ((state >= SCE_H_SGML_DEFAULT) && (state < SCE_H_SGML_BLOCK_DEFAULT)) {
 		return eScriptSGML;
 	} else if (state == SCE_H_SGML_BLOCK_DEFAULT) {
 		return eScriptSGMLblock;
@@ -94,7 +101,7 @@ static int ScriptOfState(int state) {
 	}
 }
 
-static int statePrintForState(int state, int inScriptType) {
+static int statePrintForState(int state, script_mode inScriptType) {
 	int StateToPrint;
 
 	if ((state >= SCE_HP_START) && (state <= SCE_HP_IDENTIFIER)) {
@@ -127,7 +134,7 @@ static int stateForPrintState(int StateToPrint) {
 }
 
 static inline bool IsNumber(unsigned int start, Accessor &styler) {
-	return isdigit(styler[start]) || (styler[start] == '.') ||
+	return IsADigit(styler[start]) || (styler[start] == '.') ||
 	       (styler[start] == '-') || (styler[start] == '#');
 }
 
@@ -145,6 +152,8 @@ static inline bool isStringState(int state) {
 	case SCE_HPA_STRING:
 	case SCE_HPHP_HSTRING:
 	case SCE_HPHP_SIMPLESTRING:
+	case SCE_HPHP_HSTRING_VARIABLE:
+	case SCE_HPHP_COMPLEX_VARIABLE:
 		bResult = true;
 		break;
 	default :
@@ -181,12 +190,8 @@ static void classifyAttribHTML(unsigned int start, unsigned int end, WordList &k
 	if (wordIsNumber) {
 		chAttr = SCE_H_NUMBER;
 	} else {
-		char s[30 + 1];
-		unsigned int i = 0;
-		for (; i < end - start + 1 && i < 30; i++) {
-			s[i] = static_cast<char>(tolower(styler[start + i]));
-		}
-		s[i] = '\0';
+		char s[100];
+		GetTextSegment(styler, start, end, s, sizeof(s));
 		if (keywords.InList(s))
 			chAttr = SCE_H_ATTRIBUTE;
 	}
@@ -197,16 +202,30 @@ static void classifyAttribHTML(unsigned int start, unsigned int end, WordList &k
 }
 
 static int classifyTagHTML(unsigned int start, unsigned int end,
-                           WordList &keywords, Accessor &styler) {
-	char s[30 + 1];
+                           WordList &keywords, Accessor &styler, bool &tagDontFold,
+			   bool caseSensitive) {
+	char s[30 + 2];
 	// Copy after the '<'
 	unsigned int i = 0;
 	for (unsigned int cPos = start; cPos <= end && i < 30; cPos++) {
 		char ch = styler[cPos];
-		if ((ch != '<') && (ch != '/'))
-			s[i++] = static_cast<char>(tolower(ch));
+		if ((ch != '<') && (ch != '/')) {
+			s[i++] = caseSensitive ? ch : static_cast<char>(MakeLowerCase(ch));
+		}
 	}
+
+	//The following is only a quick hack, to see if this whole thing would work
+	//we first need the tagname with a trailing space...
+	s[i] = ' ';
+	s[i+1] = '\0';
+
+	//...to find it in the list of no-container-tags
+	// (There are many more. We will need a keywordlist in the property file for this)
+	tagDontFold = (NULL != strstr("meta link img area br hr input ",s));
+
+	//now we can remove the trailing space
 	s[i] = '\0';
+
 	bool isScript = false;
 	char chAttr = SCE_H_TAGUNKNOWN;
 	if (s[0] == '!') {
@@ -220,17 +239,19 @@ static int classifyTagHTML(unsigned int start, unsigned int end,
 			isScript = 0 == strcmp(s, "script");
 		}
 	}
-	if ((chAttr == SCE_H_TAGUNKNOWN) && !keywords)
+	if ((chAttr == SCE_H_TAGUNKNOWN) && !keywords) {
 		// No keywords -> all are known
 		chAttr = SCE_H_TAG;
+		isScript = 0 == strcmp(s, "script");
+	}
 	styler.ColourTo(end, chAttr);
 	return isScript ? SCE_H_SCRIPT : chAttr;
 }
 
 static void classifyWordHTJS(unsigned int start, unsigned int end,
-                             WordList &keywords, Accessor &styler, int inScriptType) {
+                             WordList &keywords, Accessor &styler, script_mode inScriptType) {
 	char chAttr = SCE_HJ_WORD;
-	bool wordIsNumber = isdigit(styler[start]) || (styler[start] == '.');
+	bool wordIsNumber = IsADigit(styler[start]) || (styler[start] == '.');
 	if (wordIsNumber)
 		chAttr = SCE_HJ_NUMBER;
 	else {
@@ -246,18 +267,14 @@ static void classifyWordHTJS(unsigned int start, unsigned int end,
 	styler.ColourTo(end, statePrintForState(chAttr, inScriptType));
 }
 
-static int classifyWordHTVB(unsigned int start, unsigned int end, WordList &keywords, Accessor &styler, int inScriptType) {
+static int classifyWordHTVB(unsigned int start, unsigned int end, WordList &keywords, Accessor &styler, script_mode inScriptType) {
 	char chAttr = SCE_HB_IDENTIFIER;
-	bool wordIsNumber = isdigit(styler[start]) || (styler[start] == '.');
+	bool wordIsNumber = IsADigit(styler[start]) || (styler[start] == '.');
 	if (wordIsNumber)
 		chAttr = SCE_HB_NUMBER;
 	else {
-		char s[30 + 1];
-		unsigned int i = 0;
-		for (; i < end - start + 1 && i < 30; i++) {
-			s[i] = static_cast<char>(tolower(styler[start + i]));
-		}
-		s[i] = '\0';
+		char s[100];
+		GetTextSegment(styler, start, end, s, sizeof(s));
 		if (keywords.InList(s)) {
 			chAttr = SCE_HB_WORD;
 			if (strcmp(s, "rem") == 0)
@@ -271,8 +288,8 @@ static int classifyWordHTVB(unsigned int start, unsigned int end, WordList &keyw
 		return SCE_HB_DEFAULT;
 }
 
-static void classifyWordHTPy(unsigned int start, unsigned int end, WordList &keywords, Accessor &styler, char *prevWord, int inScriptType) {
-	bool wordIsNumber = isdigit(styler[start]) != 0;
+static void classifyWordHTPy(unsigned int start, unsigned int end, WordList &keywords, Accessor &styler, char *prevWord, script_mode inScriptType) {
+	bool wordIsNumber = IsADigit(styler[start]);
 	char s[30 + 1];
 	unsigned int i = 0;
 	for (; i < end - start + 1 && i < 30; i++) {
@@ -296,16 +313,12 @@ static void classifyWordHTPy(unsigned int start, unsigned int end, WordList &key
 // Called when in a PHP word
 static void classifyWordHTPHP(unsigned int start, unsigned int end, WordList &keywords, Accessor &styler) {
 	char chAttr = SCE_HPHP_DEFAULT;
-	bool wordIsNumber = isdigit(styler[start]) != 0;
+	bool wordIsNumber = IsADigit(styler[start]) || (styler[start] == '.' && start+1 <= end && IsADigit(styler[start+1]));
 	if (wordIsNumber)
 		chAttr = SCE_HPHP_NUMBER;
 	else {
-		char s[100 + 1];
-		unsigned int i = 0;
-		for (; i < end - start + 1 && i < 100; i++) {
-			s[i] = static_cast<char>(tolower(styler[start + i]));
-		}
-		s[i] = '\0';
+		char s[100];
+		GetTextSegment(styler, start, end, s, sizeof(s));
 		if (keywords.InList(s))
 			chAttr = SCE_HPHP_WORD;
 	}
@@ -333,7 +346,7 @@ static bool isWordCdata(unsigned int start, unsigned int end, Accessor &styler)
 }
 
 // Return the first state to reach when entering a scripting language
-static int StateForScript(int scriptLanguage) {
+static int StateForScript(script_type scriptLanguage) {
 	int Result;
 	switch (scriptLanguage) {
 	case eScriptVBS:
@@ -358,12 +371,22 @@ static int StateForScript(int scriptLanguage) {
 	return Result;
 }
 
-inline bool ishtmlwordchar(char ch) {
-	return isalnum(ch) || ch == '.' || ch == '-' || ch == '_' || ch == ':' || ch == '!' || ch == '#';
+static inline bool ishtmlwordchar(char ch) {
+	return !isascii(ch) ||
+		(isalnum(ch) || ch == '.' || ch == '-' || ch == '_' || ch == ':' || ch == '!' || ch == '#');
 }
 
-inline bool issgmlwordchar(char ch) {
-	return isalnum(ch) || ch == '.' || ch == '_' || ch == ':' || ch == '!' || ch == '#' || ch == '[';
+static inline bool issgmlwordchar(char ch) {
+	return !isascii(ch) ||
+		(isalnum(ch) || ch == '.' || ch == '_' || ch == ':' || ch == '!' || ch == '#' || ch == '[');
+}
+
+static inline bool IsPhpWordStart(const unsigned char ch) {
+	return (isascii(ch) && (isalpha(ch) || (ch == '_'))) || (ch >= 0x7f);
+}
+
+static inline bool IsPhpWordChar(char ch) {
+	return IsADigit(ch) || IsPhpWordStart(ch);
 }
 
 static bool InTagState(int state) {
@@ -378,6 +401,11 @@ static bool IsCommentState(const int state) {
 	return state == SCE_H_COMMENT || state == SCE_H_SGML_COMMENT;
 }
 
+static bool IsScriptCommentState(const int state) {
+	return state == SCE_HJ_COMMENT || state == SCE_HJ_COMMENTLINE || state == SCE_HJA_COMMENT ||
+		   state == SCE_HJA_COMMENTLINE || state == SCE_HB_COMMENTLINE || state == SCE_HBA_COMMENTLINE;
+}
+
 static bool isLineEnd(char ch) {
 	return ch == '\r' || ch == '\n';
 }
@@ -390,7 +418,21 @@ static bool isPHPStringState(int state) {
 	return
 	    (state == SCE_HPHP_HSTRING) ||
 	    (state == SCE_HPHP_SIMPLESTRING) ||
-	    (state == SCE_HPHP_HSTRING_VARIABLE);
+	    (state == SCE_HPHP_HSTRING_VARIABLE) ||
+	    (state == SCE_HPHP_COMPLEX_VARIABLE);
+}
+
+static int FindPhpStringDelimiter(char *phpStringDelimiter, const int phpStringDelimiterSize, int i, const int lengthDoc, Accessor &styler) {
+	int j;
+	phpStringDelimiter[0] = '\n';
+	for (j = i; j < lengthDoc && styler[j] != '\n' && styler[j] != '\r'; j++) {
+		if (j - i < phpStringDelimiterSize - 2)
+			phpStringDelimiter[j-i+1] = styler[j];
+		else
+			i++;
+	}
+	phpStringDelimiter[j-i+1] = '\0';
+	return j;
 }
 
 static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initStyle, WordList *keywordlists[],
@@ -406,6 +448,8 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 	styler.StartAt(startPos, STYLE_MAX);
 	char prevWord[200];
 	prevWord[0] = '\0';
+	char phpStringDelimiter[200]; // PHP is not limited in length, we are
+	phpStringDelimiter[0] = '\0';
 	int StateToPrint = initStyle;
 	int state = stateForPrintState(StateToPrint);
 
@@ -417,6 +461,12 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 		}
 		state = SCE_H_DEFAULT;
 	}
+	// String can be heredoc, must find a delimiter first
+	while (startPos > 0 && isPHPStringState(state) && state != SCE_HPHP_SIMPLESTRING) {
+		startPos--;
+		length++;
+		state = styler.StyleAt(startPos);
+	}
 	styler.StartAt(startPos, STYLE_MAX);
 
 	int lineCurrent = styler.GetLine(startPos);
@@ -428,18 +478,21 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 		lineState = eScriptJS << 8;
 		lineState |= styler.GetPropertyInt("asp.default.language", eScriptJS) << 4;
 	}
-	int inScriptType = (lineState >> 0) & 0x03; // 2 bits of scripting type
+	script_mode inScriptType = script_mode((lineState >> 0) & 0x03); // 2 bits of scripting mode
 	bool tagOpened = (lineState >> 2) & 0x01; // 1 bit to know if we are in an opened tag
 	bool tagClosing = (lineState >> 3) & 0x01; // 1 bit to know if we are in a closing tag
-	int aspScript = (lineState >> 4) & 0x0F; // 4 bits of script name
-	int clientScript = (lineState >> 8) & 0x0F; // 4 bits of script name
+	bool tagDontFold = false; //some HTML tags should not be folded
+	script_type aspScript = script_type((lineState >> 4) & 0x0F); // 4 bits of script name
+	script_type clientScript = script_type((lineState >> 8) & 0x0F); // 4 bits of script name
 	int beforePreProc = (lineState >> 12) & 0xFF; // 8 bits of state
 
-	int scriptLanguage = ScriptOfState(state);
+	script_type scriptLanguage = ScriptOfState(state);
 
 	const bool foldHTML = styler.GetPropertyInt("fold.html", 0) != 0;
-	const bool fold = foldHTML && styler.GetPropertyInt("fold");
+	const bool fold = foldHTML && styler.GetPropertyInt("fold", 0);
+	const bool foldHTMLPreprocessor = foldHTML && styler.GetPropertyInt("fold.html.preprocessor", 1);
 	const bool foldCompact = styler.GetPropertyInt("fold.compact", 1) != 0;
+	const bool caseSensitive = styler.GetPropertyInt("html.tags.case.sensitive", 0) != 0;
 
 	int levelPrev = styler.LevelAt(lineCurrent) & SC_FOLDLEVELNUMBERMASK;
 	int levelCurrent = levelPrev;
@@ -479,7 +532,9 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			case eScriptPHP:
 				//not currently supported				case eScriptVBS:
 
-				if ((state != SCE_HPHP_COMMENT) && (state != SCE_HPHP_COMMENTLINE) && (state != SCE_HJ_COMMENT) && (state != SCE_HJ_COMMENTLINE) && (state != SCE_HJ_COMMENTDOC)) {
+				if ((state != SCE_HPHP_COMMENT) && (state != SCE_HPHP_COMMENTLINE) && (state != SCE_HJ_COMMENT) && (state != SCE_HJ_COMMENTLINE) && (state != SCE_HJ_COMMENTDOC) && (!isStringState(state))) {
+				//Platform::DebugPrintf("state=%d, StateToPrint=%d, initStyle=%d\n", state, StateToPrint, initStyle);
+				//if ((state == SCE_HPHP_OPERATOR) || (state == SCE_HPHP_DEFAULT) || (state == SCE_HJ_SYMBOLS) || (state == SCE_HJ_START) || (state == SCE_HJ_DEFAULT)) {
 					if ((ch == '{') || (ch == '}')) {
 						levelCurrent += (ch == '{') ? 1 : -1;
 					}
@@ -511,6 +566,8 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 					}
 				}
 				break;
+			default:
+				break;
 			}
 		}
 
@@ -548,17 +605,30 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			case SCE_H_SINGLESTRING:
 			case SCE_HJ_COMMENT:
 			case SCE_HJ_COMMENTDOC:
-				// SCE_HJ_COMMENTLINE removed as this is a common thing done to hide
-				// the end of script marker from some JS interpreters.
-				//case SCE_HJ_COMMENTLINE:
+			//case SCE_HJ_COMMENTLINE: // removed as this is a common thing done to hide
+			// the end of script marker from some JS interpreters.
 			case SCE_HJ_DOUBLESTRING:
 			case SCE_HJ_SINGLESTRING:
+			case SCE_HJ_REGEX:
 			case SCE_HB_STRING:
 			case SCE_HP_STRING:
 			case SCE_HP_TRIPLE:
 			case SCE_HP_TRIPLEDOUBLE:
 				break;
 			default :
+				// check if the closing tag is a script tag
+				if (state == SCE_HJ_COMMENTLINE) {
+					char tag[7]; // room for the <script> tag
+					char chr;	// current char
+					int j=0;
+					chr = styler.SafeGetCharAt(i+2);
+					while (j < 6 && !isspacechar(chr)) {
+						tag[j++] = static_cast<char>(MakeLowerCase(chr));
+						chr = styler.SafeGetCharAt(i+2+j);
+					}
+					tag[j] = '\0';
+					if (strcmp(tag, "script") != 0) break;
+				}
 				// closing tag of the script (it's a closing HTML tag anyway)
 				styler.ColourTo(i - 1, StateToPrint);
 				state = SCE_H_TAGUNKNOWN;
@@ -578,10 +648,12 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 		         !isPHPStringState(state) &&
 		         (state != SCE_HPHP_COMMENT) &&
 		         (ch == '<') &&
-		         (chNext == '?')) {
+		         (chNext == '?') &&
+				 !IsScriptCommentState(state) ) {
+			scriptLanguage = segIsScriptingIndicator(styler, styler.GetStartSegment() + 2, i + 10, eScriptPHP);
+			if (scriptLanguage != eScriptPHP && isStringState(state)) continue;
 			styler.ColourTo(i - 1, StateToPrint);
 			beforePreProc = state;
-			scriptLanguage = segIsScriptingIndicator(styler, styler.GetStartSegment() + 2, i + 10, eScriptPHP);
 			i++;
 			visibleChars++;
 			i += PrintScriptingIndicatorOffset(styler, styler.GetStartSegment() + 2, i + 10);
@@ -595,9 +667,11 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			else
 				inScriptType = eNonHtmlPreProc;
 			// fold whole script
-			levelCurrent++;
-			if (scriptLanguage == eScriptXML)
-				levelCurrent--; // no folding of the XML first tag (all XML-like tags in this case)
+			if (foldHTMLPreprocessor){
+				levelCurrent++;
+				if (scriptLanguage == eScriptXML)
+					levelCurrent--; // no folding of the XML first tag (all XML-like tags in this case)
+			}
 			// should be better
 			ch = styler.SafeGetCharAt(i);
 			continue;
@@ -635,7 +709,8 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			scriptLanguage = eScriptVBS;
 			styler.ColourTo(i, SCE_H_ASP);
 			// fold whole script
-			levelCurrent++;
+			if (foldHTMLPreprocessor)
+				levelCurrent++;
 			// should be better
 			ch = styler.SafeGetCharAt(i);
 			continue;
@@ -644,15 +719,18 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 		/////////////////////////////////////
 		// handle the start of SGML language (DTD)
 		else if (((scriptLanguage == eScriptNone) || (scriptLanguage == eScriptXML)) &&
-		         (chPrev == '<') &&
-		         (ch == '!') &&
-		         (StateToPrint != SCE_H_CDATA) && (!IsCommentState(StateToPrint))) {
+				 (chPrev == '<') &&
+				 (ch == '!') &&
+				 (StateToPrint != SCE_H_CDATA) &&
+				 (!IsCommentState(StateToPrint)) &&
+				 (!IsScriptCommentState(StateToPrint)) ) {
 			beforePreProc = state;
 			styler.ColourTo(i - 2, StateToPrint);
 			if ((chNext == '-') && (chNext2 == '-')) {
 				state = SCE_H_COMMENT; // wait for a pending command
-			}
-			else if (isWordCdata(i + 1, i + 7, styler)) {
+				styler.ColourTo(i + 2, SCE_H_COMMENT);
+				i += 2; // follow styling after the --
+			} else if (isWordCdata(i + 1, i + 7, styler)) {
 				state = SCE_H_CDATA;
 			} else {
 				styler.ColourTo(i, SCE_H_SGML_DEFAULT); // <! is default
@@ -660,8 +738,8 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 				state = SCE_H_SGML_COMMAND; // wait for a pending command
 			}
 			// fold whole tag (-- when closing the tag)
-
-			levelCurrent++;
+			if (foldHTMLPreprocessor)
+				levelCurrent++;
 			continue;
 		}
 
@@ -718,7 +796,8 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 				inScriptType = eHtml;
 			scriptLanguage = eScriptNone;
 			// unfold all scripting languages
-			levelCurrent--;
+			if (foldHTMLPreprocessor)
+				levelCurrent--;
 			continue;
 		}
 		/////////////////////////////////////
@@ -738,8 +817,9 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			}
 			break;
 		case SCE_H_SGML_DEFAULT:
-			if (scriptLanguage == eScriptSGMLblock)
-				StateToPrint = SCE_H_SGML_BLOCK_DEFAULT;
+		case SCE_H_SGML_BLOCK_DEFAULT:
+//			if (scriptLanguage == eScriptSGMLblock)
+//				StateToPrint = SCE_H_SGML_BLOCK_DEFAULT;
 
 			if (ch == '\"') {
 				styler.ColourTo(i - 1, StateToPrint);
@@ -750,7 +830,7 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			} else if ((ch == '-') && (chPrev == '-')) {
 				styler.ColourTo(i - 2, StateToPrint);
 				state = SCE_H_SGML_COMMENT;
-			} else if (isalpha(ch) && (chPrev == '%')) {
+			} else if (isascii(ch) && isalpha(ch) && (chPrev == '%')) {
 				styler.ColourTo(i - 2, StateToPrint);
 				state = SCE_H_SGML_ENTITY;
 			} else if (ch == '#') {
@@ -759,6 +839,7 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			} else if (ch == '[') {
 				styler.ColourTo(i - 1, StateToPrint);
 				scriptLanguage = eScriptSGMLblock;
+				state = SCE_H_SGML_BLOCK_DEFAULT;
 			} else if (ch == ']') {
 				if (scriptLanguage == eScriptSGMLblock) {
 					styler.ColourTo(i, StateToPrint);
@@ -767,6 +848,7 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 					styler.ColourTo(i - 1, StateToPrint);
 					styler.ColourTo(i, SCE_H_SGML_ERROR);
 				}
+				state = SCE_H_SGML_DEFAULT;
 			} else if (scriptLanguage == eScriptSGMLblock) {
 				if ((ch == '!') && (chPrev == '<')) {
 					styler.ColourTo(i - 2, StateToPrint);
@@ -814,7 +896,11 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 				i += size - 1;
 				visibleChars += size - 1;
 				ch = styler.SafeGetCharAt(i);
-				state = SCE_H_SGML_DEFAULT;
+				if (scriptLanguage == eScriptSGMLblock) {
+					state = SCE_H_SGML_BLOCK_DEFAULT;
+				} else {
+					state = SCE_H_SGML_DEFAULT;
+				}
 				continue;
 			}
 			break;
@@ -862,7 +948,7 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			}
 			break;
 		case SCE_H_SGML_SPECIAL:
-			if (!isupper(ch)) {
+			if (!(isascii(ch) && isupper(ch))) {
 				styler.ColourTo(i - 1, StateToPrint);
 				if (isalnum(ch)) {
 					state = SCE_H_SGML_ERROR;
@@ -875,7 +961,7 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			if (ch == ';') {
 				styler.ColourTo(i, StateToPrint);
 				state = SCE_H_SGML_DEFAULT;
-			} else if (!isalnum(ch) && ch != '-' && ch != '.') {
+			} else if (!(isascii(ch) && isalnum(ch)) && ch != '-' && ch != '.') {
 				styler.ColourTo(i, SCE_H_SGML_ERROR);
 				state = SCE_H_SGML_DEFAULT;
 			}
@@ -885,14 +971,16 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 				styler.ColourTo(i, StateToPrint);
 				state = SCE_H_DEFAULT;
 			}
-			if (ch != '#' && !isalnum(ch)) {	// Should check that '#' follows '&', but it is unlikely anyway...
+			if (ch != '#' && !(isascii(ch) && isalnum(ch))	// Should check that '#' follows '&', but it is unlikely anyway...
+				&& ch != '.' && ch != '-' && ch != '_' && ch != ':') { // valid in XML
 				styler.ColourTo(i, SCE_H_TAGUNKNOWN);
 				state = SCE_H_DEFAULT;
 			}
 			break;
 		case SCE_H_TAGUNKNOWN:
 			if (!ishtmlwordchar(ch) && !((ch == '/') && (chPrev == '<')) && ch != '[') {
-				int eClass = classifyTagHTML(styler.GetStartSegment(), i - 1, keywords, styler);
+				int eClass = classifyTagHTML(styler.GetStartSegment(),
+					i - 1, keywords, styler, tagDontFold, caseSensitive);
 				if (eClass == SCE_H_SCRIPT) {
 					if (!tagClosing) {
 						inScriptType = eNonHtmlScript;
@@ -911,10 +999,12 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 						state = SCE_H_DEFAULT;
 					}
 					tagOpened = false;
-					if (tagClosing) {
-						levelCurrent--;
-					} else {
-						levelCurrent++;
+					if (!tagDontFold){
+						if (tagClosing) {
+							levelCurrent--;
+						} else {
+							levelCurrent++;
+						}
 					}
 					tagClosing = false;
 				} else if (ch == '/' && chNext == '>') {
@@ -957,10 +1047,13 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 						state = SCE_H_DEFAULT;
 					}
 					tagOpened = false;
-					if (tagClosing)
-						levelCurrent--;
-					else
-						levelCurrent++;
+					if (!tagDontFold){
+						if (tagClosing){
+							levelCurrent--;
+						} else {
+							levelCurrent++;
+						}
+					}
 					tagClosing = false;
 				} else if (ch == '=') {
 					styler.ColourTo(i, SCE_H_OTHER);
@@ -980,10 +1073,13 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 					state = SCE_H_DEFAULT;
 				}
 				tagOpened = false;
-				if (tagClosing)
-					levelCurrent--;
-				else
-					levelCurrent++;
+				if (!tagDontFold){
+					if (tagClosing){
+						levelCurrent--;
+					} else {
+						levelCurrent++;
+					}
+				}
 				tagClosing = false;
 			} else if (ch == '\"') {
 				styler.ColourTo(i - 1, StateToPrint);
@@ -1032,10 +1128,10 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			break;
 		case SCE_H_VALUE:
 			if (!ishtmlwordchar(ch)) {
-				if (ch == '\"') {
+				if (ch == '\"' && chPrev == '=') {
 					// Should really test for being first character
 					state = SCE_H_DOUBLESTRING;
-				} else if (ch == '\'') {
+				} else if (ch == '\'' && chPrev == '=') {
 					state = SCE_H_SINGLESTRING;
 				} else {
 					if (IsNumber(styler.GetStartSegment(), styler)) {
@@ -1051,10 +1147,13 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 							state = SCE_H_DEFAULT;
 						}
 						tagOpened = false;
-						if (tagClosing)
-							levelCurrent--;
-						else
-							levelCurrent++;
+						if (!tagDontFold){
+							if (tagClosing){
+								levelCurrent--;
+							} else {
+								levelCurrent++;
+							}
+						}
 						tagClosing = false;
 					} else {
 						state = SCE_H_OTHER;
@@ -1385,7 +1484,7 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			break;
 			///////////// start - PHP state handling
 		case SCE_HPHP_WORD:
-			if (!iswordstart(ch)) {
+			if (!iswordchar(ch)) {
 				classifyWordHTPHP(styler.GetStartSegment(), i - 1, keywords5, styler);
 				if (ch == '/' && chNext == '*') {
 					i++;
@@ -1397,9 +1496,13 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 					state = SCE_HPHP_COMMENTLINE;
 				} else if (ch == '\"') {
 					state = SCE_HPHP_HSTRING;
+					strcpy(phpStringDelimiter, "\"");
+				} else if (styler.Match(i, "<<<")) {
+					state = SCE_HPHP_HSTRING;
+					i = FindPhpStringDelimiter(phpStringDelimiter, sizeof(phpStringDelimiter), i + 3, lengthDoc, styler);
 				} else if (ch == '\'') {
 					state = SCE_HPHP_SIMPLESTRING;
-				} else if (ch == '$') {
+				} else if (ch == '$' && IsPhpWordStart(chNext)) {
 					state = SCE_HPHP_VARIABLE;
 				} else if (isoperator(ch)) {
 					state = SCE_HPHP_OPERATOR;
@@ -1409,7 +1512,10 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			}
 			break;
 		case SCE_HPHP_NUMBER:
-			if (!isdigit(ch)) {
+			// recognize bases 8,10 or 16 integers OR floating-point numbers
+			if (!IsADigit(ch)
+				&& strchr(".xXabcdefABCDEF", ch) == NULL
+				&& ((ch != '-' && ch != '+') || (chPrev != 'e' && chPrev != 'E'))) {
 				styler.ColourTo(i - 1, SCE_HPHP_NUMBER);
 				if (isoperator(ch))
 					state = SCE_HPHP_OPERATOR;
@@ -1418,7 +1524,7 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			}
 			break;
 		case SCE_HPHP_VARIABLE:
-			if (!iswordstart(ch)) {
+			if (!IsPhpWordChar(ch)) {
 				styler.ColourTo(i - 1, SCE_HPHP_VARIABLE);
 				if (isoperator(ch))
 					state = SCE_HPHP_OPERATOR;
@@ -1439,13 +1545,19 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			}
 			break;
 		case SCE_HPHP_HSTRING:
-			if (ch == '\\') {
+			if (ch == '\\' && (phpStringDelimiter[0] == '\"' || chNext == '$' || chNext == '{')) {
 				// skip the next char
 				i++;
-			} else if (ch == '$') {
+			} else if (((ch == '{' && chNext == '$') || (ch == '$' && chNext == '{'))
+				&& IsPhpWordStart(chNext2)) {
+				styler.ColourTo(i - 1, StateToPrint);
+				state = SCE_HPHP_COMPLEX_VARIABLE;
+			} else if (ch == '$' && IsPhpWordStart(chNext)) {
 				styler.ColourTo(i - 1, StateToPrint);
 				state = SCE_HPHP_HSTRING_VARIABLE;
-			} else if (ch == '\"') {
+			} else if (styler.Match(i, phpStringDelimiter)) {
+				if (strlen(phpStringDelimiter) > 1)
+					i += strlen(phpStringDelimiter) - 1;
 				styler.ColourTo(i, StateToPrint);
 				state = SCE_HPHP_DEFAULT;
 			}
@@ -1460,16 +1572,22 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 			}
 			break;
 		case SCE_HPHP_HSTRING_VARIABLE:
-			if (!iswordstart(ch)) {
+			if (!IsPhpWordChar(ch)) {
 				styler.ColourTo(i - 1, StateToPrint);
 				i--; // strange but it works
 				state = SCE_HPHP_HSTRING;
 			}
 			break;
+		case SCE_HPHP_COMPLEX_VARIABLE:
+			if (ch == '}') {
+				styler.ColourTo(i, StateToPrint);
+				state = SCE_HPHP_HSTRING;
+			}
+			break;
 		case SCE_HPHP_OPERATOR:
 		case SCE_HPHP_DEFAULT:
 			styler.ColourTo(i - 1, StateToPrint);
-			if (isdigit(ch)) {
+			if (IsADigit(ch) || (ch == '.' && IsADigit(chNext))) {
 				state = SCE_HPHP_NUMBER;
 			} else if (iswordstart(ch)) {
 				state = SCE_HPHP_WORD;
@@ -1483,9 +1601,13 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 				state = SCE_HPHP_COMMENTLINE;
 			} else if (ch == '\"') {
 				state = SCE_HPHP_HSTRING;
+				strcpy(phpStringDelimiter, "\"");
+			} else if (styler.Match(i, "<<<")) {
+				state = SCE_HPHP_HSTRING;
+				i = FindPhpStringDelimiter(phpStringDelimiter, sizeof(phpStringDelimiter), i + 3, lengthDoc, styler);
 			} else if (ch == '\'') {
 				state = SCE_HPHP_SIMPLESTRING;
-			} else if (ch == '$') {
+			} else if (ch == '$' && IsPhpWordStart(chNext)) {
 				state = SCE_HPHP_VARIABLE;
 			} else if (isoperator(ch)) {
 				state = SCE_HPHP_OPERATOR;
@@ -1541,7 +1663,7 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 	}
 
 	StateToPrint = statePrintForState(state, inScriptType);
-	styler.ColourTo(lengthDoc - 1, StateToPrint);
+		styler.ColourTo(lengthDoc - 1, StateToPrint);
 
 	// Fill in the real level of the next line, keeping the current flags as they will be filled in later
 	if (fold) {
@@ -1551,7 +1673,7 @@ static void ColouriseHyperTextDoc(unsigned int startPos, int length, int initSty
 }
 
 static bool isASPScript(int state) {
-	return 
+	return
 		(state >= SCE_HJA_START && state <= SCE_HJA_REGEX) ||
 		(state >= SCE_HBA_START && state <= SCE_HBA_STRINGEOL) ||
 		(state >= SCE_HPA_DEFAULT && state <= SCE_HPA_IDENTIFIER);
@@ -1618,7 +1740,8 @@ static void ColouriseHTMLPiece(StyleContext &sc, WordList *keywordlists[]) {
 	} else if (sc.state == SCE_H_ENTITY) {
 		if (sc.ch == ';') {
 			sc.ForwardSetState(SCE_H_DEFAULT);
-		} else if (sc.ch != '#' && (sc.ch < 0x80) && !isalnum(sc.ch)) {	// Should check that '#' follows '&', but it is unlikely anyway...
+		} else if (sc.ch != '#' && (sc.ch < 0x80) && !isalnum(sc.ch)	// Should check that '#' follows '&', but it is unlikely anyway...
+			&& sc.ch != '.' && sc.ch != '-' && sc.ch != '_' && sc.ch != ':') { // valid in XML
 			sc.ChangeState(SCE_H_TAGUNKNOWN);
 			sc.SetState(SCE_H_DEFAULT);
 		}
@@ -1678,7 +1801,7 @@ static void ColouriseHTMLPiece(StyleContext &sc, WordList *keywordlists[]) {
 			sc.SetState(SCE_H_OTHER);
 		}
 	}
-	
+
 	if (sc.state == SCE_H_DEFAULT) {
 		if (sc.ch == '<') {
 			if (sc.Match("<!--"))
@@ -1689,9 +1812,9 @@ static void ColouriseHTMLPiece(StyleContext &sc, WordList *keywordlists[]) {
 			sc.SetState(SCE_H_ENTITY);
 		}
 	} else if ((sc.state == SCE_H_OTHER) || (sc.state == SCE_H_VALUE)) {
-		if (sc.ch == '\"') {
+		if (sc.ch == '\"' && sc.chPrev == '=') {
 			sc.SetState(SCE_H_DOUBLESTRING);
-		} else if (sc.ch == '\'') {
+		} else if (sc.ch == '\'' && sc.chPrev == '=') {
 			sc.SetState(SCE_H_SINGLESTRING);
 		} else if (IsADigit(sc.ch)) {
 			sc.SetState(SCE_H_NUMBER);
@@ -1710,15 +1833,15 @@ static void ColouriseASPPiece(StyleContext &sc, WordList *keywordlists[]) {
 		sc.SetState(SCE_H_ASP);
 		sc.Forward();
 		sc.ForwardSetState(SCE_H_DEFAULT);
-	} 
-			 
+	}
+
 	// Handle some ASP script
 	if (sc.state >= SCE_HBA_START && sc.state <= SCE_HBA_STRINGEOL) {
 		ColouriseHBAPiece(sc, keywordlists);
 	} else if (sc.state >= SCE_H_DEFAULT && sc.state <= SCE_H_SGML_BLOCK_DEFAULT) {
 		ColouriseHTMLPiece(sc, keywordlists);
 	}
-	
+
 	// Enter new sc.state
 	if ((sc.state == SCE_H_DEFAULT) || (sc.state == SCE_H_TAGUNKNOWN)) {
 		if (sc.Match('<', '%')) {
@@ -1753,22 +1876,22 @@ static void ColouriseASPDoc(unsigned int startPos, int length, int initStyle, Wo
 static void ColourisePHPPiece(StyleContext &sc, WordList *keywordlists[]) {
 	// Possibly exit current state to either SCE_H_DEFAULT or SCE_HBA_DEFAULT
 	if (sc.state >= SCE_HPHP_DEFAULT && sc.state <= SCE_HPHP_OPERATOR) {
-		if (!isPHPStringState(sc.state) && 
+		if (!isPHPStringState(sc.state) &&
 			(sc.state != SCE_HPHP_COMMENT) &&
 			(sc.Match('?', '>'))) {
 			sc.SetState(SCE_H_QUESTION);
 			sc.Forward();
 			sc.ForwardSetState(SCE_H_DEFAULT);
 		}
-	} 
-			 
+	}
+
 	if (sc.state >= SCE_H_DEFAULT && sc.state <= SCE_H_SGML_BLOCK_DEFAULT) {
 		ColouriseHTMLPiece(sc, keywordlists);
 	}
-	
+
 	// Handle some PHP script
 	if (sc.state == SCE_HPHP_WORD) {
-		if (!IsAWordStart(sc.ch)) {
+		if (!IsPhpWordChar(static_cast<char>(sc.ch))) {
 			sc.SetState(SCE_HPHP_DEFAULT);
 		}
 	} else if (sc.state == SCE_HPHP_COMMENTLINE) {
@@ -1790,13 +1913,13 @@ static void ColourisePHPPiece(StyleContext &sc, WordList *keywordlists[]) {
 			sc.ForwardSetState(SCE_HPHP_DEFAULT);
 		}
 	} else if (sc.state == SCE_HPHP_VARIABLE) {
-		if (!IsAWordStart(sc.ch)) {
+		if (!IsPhpWordChar(static_cast<char>(sc.ch))) {
 			sc.SetState(SCE_HPHP_DEFAULT);
 		}
 	} else if (sc.state == SCE_HPHP_OPERATOR) {
 		sc.SetState(SCE_HPHP_DEFAULT);
 	}
-	
+
 	// Enter new sc.state
 	if ((sc.state == SCE_H_DEFAULT) || (sc.state == SCE_H_TAGUNKNOWN)) {
 		if (sc.Match("<?php")) {
@@ -1810,7 +1933,7 @@ static void ColourisePHPPiece(StyleContext &sc, WordList *keywordlists[]) {
 		}
 	}
 	if (sc.state == SCE_HPHP_DEFAULT) {
-		if (IsAWordStart(sc.ch)) {
+		if (IsPhpWordStart(static_cast<char>(sc.ch))) {
 			sc.SetState(SCE_HPHP_WORD);
 		} else if (sc.ch == '#') {
 			sc.SetState(SCE_HPHP_COMMENTLINE);
@@ -1824,7 +1947,7 @@ static void ColourisePHPPiece(StyleContext &sc, WordList *keywordlists[]) {
 			sc.SetState(SCE_HPHP_HSTRING);
 		} else if (sc.ch == '\'') {
 			sc.SetState(SCE_HPHP_SIMPLESTRING);
-		} else if (sc.ch == '$') {
+		} else if (sc.ch == '$' && IsPhpWordStart(static_cast<char>(sc.chNext))) {
 			sc.SetState(SCE_HPHP_VARIABLE);
 		} else if (isoperator(static_cast<char>(sc.ch))) {
 			sc.SetState(SCE_HPHP_OPERATOR);
@@ -1842,7 +1965,35 @@ static void ColourisePHPDoc(unsigned int startPos, int length, int initStyle, Wo
 	sc.Complete();
 }
 
-LexerModule lmHTML(SCLEX_HTML, ColouriseHyperTextDoc, "hypertext");
-LexerModule lmXML(SCLEX_XML, ColouriseHyperTextDoc, "xml");
-LexerModule lmASP(SCLEX_ASP, ColouriseASPDoc, "asp");
-LexerModule lmPHP(SCLEX_PHP, ColourisePHPDoc, "php");
+static void ColourisePHPScriptDoc(unsigned int startPos, int length, int initStyle, WordList *keywordlists[],
+                                               Accessor &styler) {
+	if(startPos == 0) initStyle = SCE_HPHP_DEFAULT;
+		ColouriseHyperTextDoc(startPos,length,initStyle,keywordlists,styler);
+}
+
+static const char * const htmlWordListDesc[] = {
+	"HTML elements and attributes",
+	"JavaScript keywords",
+	"VBScript keywords",
+	"Python keywords",
+	"PHP keywords",
+	"SGML and DTD keywords",
+	0,
+};
+
+static const char * const phpscriptWordListDesc[] = {
+	"", //Unused
+	"", //Unused
+	"", //Unused
+	"", //Unused
+	"PHP keywords",
+	"", //Unused
+	0,
+};
+
+LexerModule lmHTML(SCLEX_HTML, ColouriseHyperTextDoc, "hypertext", 0, htmlWordListDesc);
+LexerModule lmXML(SCLEX_XML, ColouriseHyperTextDoc, "xml", 0, htmlWordListDesc);
+// SCLEX_ASP and SCLEX_PHP should not be used in new code: use SCLEX_HTML instead.
+LexerModule lmASP(SCLEX_ASP, ColouriseASPDoc, "asp", 0, htmlWordListDesc);
+LexerModule lmPHP(SCLEX_PHP, ColourisePHPDoc, "php", 0, htmlWordListDesc);
+LexerModule lmPHPSCRIPT(SCLEX_PHPSCRIPT, ColourisePHPScriptDoc, "phpscript", 0, phpscriptWordListDesc);
