@@ -77,6 +77,8 @@
 #define IDM_COMPLETEWORD	234
 #define IDM_EXPAND			235
 #define IDM_TOGGLE_FOLDALL	236
+#define IDM_TOGGLE_FOLDRECURSIVE 237
+#define IDM_EXPAND_ENSURECHILDRENVISIBLE 238
 #define IDM_UPRCASE			240
 #define IDM_LWRCASE			241
 #define IDM_ABBREV			242
