@@ -376,7 +376,7 @@ enum IndentationStatus {
 	isKeyWordStart	// Keywords that cause indentation
 };
 
-int IntFromHexDigit(const char ch);
+int IntFromHexDigit(int ch);
 int IntFromHexByte(const char *hexByte);
 
 class StyleDefinition {
@@ -806,6 +806,8 @@ protected:
 	void Expand(int &line, bool doExpand, bool force = false,
 	            int visLevels = 0, int level = -1);
 	void FoldAll();
+	void ToggleFoldRecursive(int line, int level);
+	void EnsureAllChildrenVisible(int line, int level);
 	void EnsureRangeVisible(int posStart, int posEnd, bool enforcePolicy=true);
 	void GotoLineEnsureVisible(int line);
 	bool MarginClick(int position, int modifiers);
@@ -956,7 +958,6 @@ time_t GetModTime(const char *fullPath);
 bool IsUntitledFileName(const char *name);
 void LowerCaseString(char *s);
 void ChopTerminalSlash(char *path);
-int IntFromHexDigit(const char ch);
 long ColourOfProperty(PropSet &props, const char *key, ColourDesired colourDefault);
 char *Slash(const char *s);
 unsigned int UnSlash(char *s);
