@@ -323,8 +323,8 @@ void SciTEBase::ReadLocalPropFile() {
 	props.Set("ChromeHighlight", "#FFFFFF");
 }
 
-int IntFromHexDigit(const char ch) {
-	if (isdigit(ch)) {
+int IntFromHexDigit(int ch) {
+	if ((ch >= '0') && (ch <= '9')) {
 		return ch - '0';
 	} else if (ch >= 'A' && ch <= 'F') {
 		return ch - 'A' + 10;
@@ -549,7 +549,9 @@ void SciTEBase::SetStyleFor(Window &win, const char *lang) {
 
 void LowerCaseString(char *s) {
 	while (*s) {
-		*s = static_cast<char>(tolower(*s));
+		if ((*s >= 'A') && (*s <= 'Z')) {
+			*s = static_cast<char>(*s - 'A' + 'a');
+		}
 		s++;
 	}
 }
@@ -651,11 +653,12 @@ SString SciTEBase::FindLanguageProperty(const char *pattern, const char *default
 static const char *propertiesToForward[] = {
 	"fold",
 	"fold.comment",
-	"fold.comment.python",
 	"fold.compact",
 	"fold.at.else",
 	"fold.html",
 	"fold.html.preprocessor",
+	"fold.perl.package",
+	"fold.perl.pod",
 	"fold.preprocessor",
 	"fold.quotes.python",
 	"fold.verilog.flags",
@@ -746,7 +749,7 @@ void SciTEBase::ReadProperties() {
 	SendEditor(SCI_SETCODEPAGE, codePage);
 	SendOutput(SCI_SETCODEPAGE, codePage);
 
-	characterSet = props.GetInt("character.set");
+	characterSet = props.GetInt("character.set", SC_CHARSET_DEFAULT);
 
 #ifdef unix
 	SString localeCType = props.Get("LC_CTYPE");
@@ -1500,6 +1503,8 @@ int SciTEBase::GetMenuCommandAsInt(SString commandName) {
 		{"IDM_COMPLETEWORD",	IDM_COMPLETEWORD},
 		{"IDM_EXPAND",	IDM_EXPAND},
 		{"IDM_TOGGLE_FOLDALL",	IDM_TOGGLE_FOLDALL},
+		{"IDM_TOGGLE_FOLDRECURSIVE",	IDM_TOGGLE_FOLDRECURSIVE},
+		{"IDM_EXPAND_ENSURECHILDRENVISIBLE",	IDM_EXPAND_ENSURECHILDRENVISIBLE},
 		{"IDM_UPRCASE",	IDM_UPRCASE},
 		{"IDM_LWRCASE",	IDM_LWRCASE},
 		{"IDM_ABBREV",	IDM_ABBREV},
